# -*- coding: utf-8 -*-

import os
import re
import json
import time
import uuid
import random
import asyncio
import logging
import shelve

from typing import Optional, Dict, Tuple, Any, Union
from playwright.async_api import BrowserContext as AsyncContext
from playwright.async_api import Page as AsyncPage
from cf_clearance import StealthConfig


fmt='%(asctime)s %(levelname)s %(module)s %(funcName)s %(lineno)d %(message)s'
logging.basicConfig(filename='logfile.log', level=logging.INFO,
                    format=fmt)
formatter = logging.Formatter(fmt=fmt)
handler = logging.StreamHandler()
handler.setFormatter(formatter)

def get_logger(name):
    logger = logging.getLogger(name)
    logger.addHandler(handler)
    return logger

logger = get_logger(__name__)

async def async_stealth(
    page_or_context: Union[AsyncContext, AsyncPage],
    config: StealthConfig = None,
    pure: bool = True,
):
    """teaches asynchronous playwright Page to be stealthy like a ninja!"""
    for script in (config or StealthConfig()).enabled_scripts:
        await page_or_context.add_init_script(script)
    if pure:
        await page_or_context.route(
            re.compile(
                r"(.*\.png(\?.*|$))|(.*\.jpg(\?.*|$))|(.*\.jpeg(\?.*|$))|(.*\.css(\?.*|$))"
            ),
            lambda route: route.abort("blockedbyclient"),
        )

class ChatGPTException(Exception):
    def __init__(self, error):
        super().__init__(self)
        self.error = error

    def __repr__(self):
        return self.error

    def __str__(self):
        return self.error

class TooManyRequestsException(ChatGPTException):
    pass

auto_send_messages = [
    "Hello",
    "Hi",
    "Hey",
    "How are you?",
    "What's up?",
    "Yo",
]

class MessageParser:
    def __init__(self):
        self.pos = 0
        self.message = None
        self.start = time.time()

    def feed(self, message: str):
        self.message = message

    def get_message(self):
        if time.time() - self.start < 1.0:
            return None
        try:
            last_index = self.message.rindex('\n\n', self.pos)
            pos = self.pos
            self.pos = last_index + 2
            self.start = time.time()
            return self.message[pos: self.pos]
        except ValueError:
            return None

    def get_remanent_message(self):
        return self.message[self.pos:]

class ChatGPTUser:

    def __init__(self, user_id: str):
        self.user_id = user_id
        self.conversation_id = None
        self.parent_message_id = None
        self.expiration = time.time() + 15*60.0

    def reset_expiration(self):
        self.expiration = time.time() + 15*60.0

    def is_expired(self):
        return self.expiration < time.time()

class ChatGPTBot:

    def __init__(self, PLAY: Any, user: str, password: str):
        self.page: Optional[Any] = None
        self.access_token: Optional[str] = None

        self.PLAY = PLAY
        self.user = user
        self.password = password
        self._standby: bool = False
        self._busy: bool = False

        self.lock = asyncio.Lock()

        self.conversation_id = None
        self.parent_message_id = None

        if not os.path.exists('.db'):
            os.mkdir('.db')
        self.users = shelve.open(f".db/{user}-1")
        self.expired_user = shelve.open(f".db/{user}-2")

        self.alive_counter = 0

    @property
    def standby(self):
        return self._standby

    @standby.setter
    def standby(self, value):
        self._standby = value

    @property
    def busy(self):
        return self._busy

    @busy.setter
    def busy(self, value):
        self._busy = value

    def handle_expired_user(self, user: ChatGPTUser):
        self.expired_user[user.user_id] = user
        self.users.pop(user.user_id, None)

    def reset_alive_counter(self):
        self.alive_counter = 0

    async def heart_beat(self):
        try:
            while True:
                try:
                    await self.check_expiration()
                except Exception as e:
                    logger.exception(e)

                try:
                    await self.keep_alive()
                except Exception as e:
                    logger.exception(e)

                await asyncio.sleep(1.0)
        except asyncio.CancelledError:
            logger.info("+++++++saving data on exit...")
            self.close()

    def close(self):
        self.users.close()
        self.expired_user.close()

    async def check_expiration(self):
        for user in self.users.values():
            if user.is_expired():
                self.handle_expired_user(user)

    async def keep_alive(self):
        self.alive_counter += 1
        if self.alive_counter < 15 * 60:
            return
        self.reset_alive_counter()
        try:
            logger.info("keep alive")
            kv = '{"event":"session","data":{"trigger":"getSession"},"timestamp": %s}' % (round(time.time()), )
            script = '''
            async ({kv}) => {
                window.localStorage.setItem('nextauth.message', kv)
            }
            '''
            ret = await self.page.evaluate(script, {'kv': kv})
            # window.localStorage.setItem('nextauth.message', '{"event":"session","data":{"trigger":"getSession"},"timestamp":'1672014120'}')
            # async with self.lock:
            #     await self.page.goto("https://chat.openai.com/chat", timeout=60*1000)
            msg = auto_send_messages[random.randint(0, len(auto_send_messages) - 1)]
            async for msg in self.send_message('main', msg):
                logger.info(msg)
            self.standby = False
        except Exception as e:
            logger.exception(e)
            await self.reload()

    async def on_response(self, response):
        url = response.url
        if not (url.endswith('.otf') or url.endswith('.js')):
            logger.info(f"Response: {url} {response.status}")
        if url.endswith("api/auth/session"):
            if response.status == 200:
                body = await response.json()
                logger.info(f"body: {body}")
                if body:
                    self.access_token = body["accessToken"]
                    logger.info("++++=access token: %s", self.access_token)

    async def init(self):
        asyncio.create_task(self.heart_beat())

        BROWSER = await self.PLAY.firefox.launch_persistent_context(
            user_data_dir=f"/tmp/playwright/firefox-{self.user}",
            headless=False
        )

        self.page = await BROWSER.new_page()
        self.page.on('response', self.on_response)

        await async_stealth(self.page, pure=False)

        try:
            await asyncio.sleep(2.0)
            await self.page.goto("https://chat.openai.com/chat", timeout=60*1000)

            while True:
                if self.access_token:
                    return
                buttons = await self.page.query_selector_all("button[class*='btn flex justify-center gap-2 btn-primary']")
                if buttons:
                    break
                if await self.page.query_selector("textarea"):
                    logger.info("++++get_access_token")
                    if await self.get_access_token():
                        return
                await asyncio.sleep(3.0)
                logger.info("waiting for page...")

            await self.page.locator("button", has_text="Log in").click(timeout=1*1000)

            username = self.page.locator('input[name="username"]')
            await username.fill(self.user)
            await username.press("Enter")

            password = self.page.locator('input[name="password"]')
            await password.fill(self.password, timeout=600000)
            await password.press("Enter")
            
            # On first login
            # next_button = self.page.locator("button", has_text="Next")
            # await next_button.click()
            # next_button = self.page.locator("button", has_text="Next")
            # await next_button.click()
            # next_button = self.page.locator("button", has_text="Done")
            # await next_button.click()
        except Exception as e:
            logger.exception(e)

        while not self.access_token:
            logger.info("waiting for access token")
            await asyncio.sleep(2.0)

    async def get_access_token(self):
        if self.access_token:
            return True
        try:
            script = '''
            async ({}) => {
                console.log("fetch access token");
                const res = await fetch('https://chat.openai.com/api/auth/session', {method: 'GET'});
                console.log(res);
                if (!res.ok) {
                    console.log("++++++++status text:", res.statusText);
                    return {status: false, result: await res.statusText};
                }
                return {status: true, result: await res.json()};
            }
        '''
            ret = await self.page.evaluate(script, {})
            logger.info("++++++ret: %s", ret)
            if not ret['status']:
                return False
            if 'error' in ret and ret['error']:
                return False
            result = ret['result']
            if not result:
                return False
            self.access_token = result['accessToken']
            return True
        except Exception as e:
            logger.exception(e)
        return False

    async def reload(self):
        await self.page.reload()

    def get_user(self, user_id):
        if user_id in self.users:
            user = self.users[user_id]
            user.reset_expiration()
        elif user_id in self.expired_user:
            user = self.expired_user[user_id]
            user.reset_expiration()
            del self.expired_user[user_id]
            self.users[user_id] = user
        else:
            user = ChatGPTUser(user_id)
            self.users[user_id] = user        
        return user

    def reset_conversation_id(self, user_id):
        assert user_id in self.users
        user = self.users[user_id]
        user.conversation_id = None
        self.users[user_id] = user

    async def send_message(self, user_id, message):
        try:
            async with self.lock:
                user = self.get_user(user_id)
                self.busy = True
                async for msg in self._send_message(user, message):
                    yield msg
        finally:
            self.busy = False
        return

    async def _send_message(self, user, message):
        message_id = str(uuid.uuid4())
        if not user.parent_message_id:
            user.parent_message_id = str(uuid.uuid4())

        parser = MessageParser()

        logger.info("++++++conversation_id: %s, parent_message_id: %s", user.conversation_id, user.parent_message_id)
        body = {
            "action": "next",
            "messages": [
                {
                    "id": message_id,
                    "role": 'user',
                    "content": {
                        "content_type": 'text',
                        "parts": [message]
                    }
                }
            ],
            "model": 'text-davinci-002-render',
            "parent_message_id": user.parent_message_id
        }

        if user.conversation_id:
            body['conversation_id'] = user.conversation_id

        url = "https://chat.openai.com/backend-api/conversation"

        script = '''
        async ({ url, body, accessToken }) => {
            console.log(url, body, accessToken);
            const res = await fetch(url, {
            method: 'POST',
            body: body,
            signal: null,
            headers: {
                accept: 'text/event-stream',
                'x-openai-assistant-app-id': '',
                authorization: `Bearer ${accessToken}`,
                'content-type': 'application/json'
            }
            });
            console.log(res);
            if (!res.ok) {
                console.log("++++++++status text:", res.statusText);
                return res.text();
            }

            const reader = res.body.getReader()
            window.reader = reader
            window.result = res;
            return "OK";
            return await res.text();
        }
    '''
        ret = await self.page.evaluate(script, { 'url': url, 'body': json.dumps(body), "accessToken": self.access_token })
        logger.info("+++++++++ret: %s", ret)
        if not ret == "OK":
            await self.reload()

            if 'Conversation not found' in ret:
                self.reset_conversation_id(user.user_id)
            elif 'Rate limit reached' in ret:
                self.standby = True
            elif 'Too many requests' in ret:
                self.standby = True
                raise TooManyRequestsException(ret)
            raise ChatGPTException(ret)
        yield "[BEGIN]\n"
        done = False
        buffer = b''
        while not done:
            ret = await self.page.evaluate("""async () => {
                const { done, value } = await window.reader.read();
                // console.log(done, value);
                const hexValue = Array.from(value).map((i) => ('0' + i.toString(16)).slice(-2)).join('');
                return hexValue;
            }
            """, {})
            buffer += bytes.fromhex(ret)
            messages = buffer.split(b'\n\n')
            buffer = messages[-1]
            messages = messages[:-1]
            for msg in messages:
                if not msg:
                    continue
                if msg.find(b'\r\n'):
                    # ignore data like b'event: ping\r\ndata: 2023-01-11 03:20:19.692082\r\n\r\n
                    msg = msg.split(b'\r\n')[-1]
                msg = msg[len(b'data: '):]
                if b'[DONE]' in msg:
                    done = True
                    break
                msg = json.loads(msg)
                user.conversation_id = msg['conversation_id']
                user.parent_message_id = msg['message']['id']
                parser.feed(msg['message']['content']['parts'][0])
                message = parser.get_message()
                if message:
                    logger.info("++++++message %s", message)
                    yield message
                else:
                    await asyncio.sleep(0.0)
        message = parser.get_remanent_message()
        if message:
            logger.info("++++++last message: %s", message)
            yield message
        self.reset_alive_counter()
        self.users[user.user_id] = user
        return

async def run():
    bot = ChatGPTBot('usr', 'psw')
    await bot.init()
    question = "hello, will openai do evil things?"
    question = 'write a http server in python, demand: 1. use fastapi and uvicorn 2. can process post and get request 3. handle unexpected exception'
    async for msg in bot.send_message(question):
        print(msg)

if __name__ == "__main__":
    asyncio.run(run())



import os
import time
import json
import uuid
import asyncio
from typing import List, Dict, Tuple, Optional, Any
import shelve
from datetime import datetime
from dataclasses import dataclass
import openai
from pymixin import log

logger = log.get_logger(__name__)
logger.addHandler(log.handler)

@dataclass
class Message:
    message: str
    parent_message_id: Optional[str]
    completion: str

if not os.path.exists('.db'):
    os.mkdir('.db')
g_conversations = shelve.open(f".db/conversations")

class ChatGPTBot:
    def __init__(self, api_key: str, stream=True):
        openai.api_key = api_key
        self.conversation_id = uuid.uuid4()

        self.standby = False
        self.users: Dict[str, bool] = {}

        self.lock = asyncio.Lock()
        self.stream = stream

    async def init(self):
        pass
    
    def generate_key(self, conversation_id: str, message_id: str):
        return f'{conversation_id}-{message_id}'

    def get_parent_messsage(self, conversation_id: str, message_id: str) -> Optional[Message]:
        key = self.generate_key(conversation_id, message_id)
        if key in g_conversations:
            return g_conversations[key]
        return None

    def add_messsage(self, conversation_id: str, query: str, reply: str) -> str:

        message_id = str(uuid.uuid4())
        key = self.generate_key(conversation_id, message_id)
        assert not key in g_conversations

        parent_message_id = self.get_last_message_id(conversation_id)
        message = Message(query, parent_message_id, reply)
        g_conversations[key] = message
        self.save_last_message_id(conversation_id, message_id)
        return message_id

    def get_last_message_id(self, conversation_id: str) -> Optional[str]:
        key = f'{conversation_id}-last_message_id'
        if key in g_conversations:
            return g_conversations[key]
        return None

    def save_last_message_id(self, conversation_id: str, message_id):
        key = f'{conversation_id}-last_message_id'
        g_conversations[key] = message_id

    def generate_prompt(self, conversation_id: str, message: str) -> List[Dict[str, str]]:
        parent_message_id = self.get_last_message_id(conversation_id)
        parent_messages: list[Message] = []
        total_length = 0
        context_messages=[
            {"role": "system", "content": "You are a helpful assistant"},
        ]
        if not parent_message_id:
            context_messages.append({"role": "user", "content": message})
            return context_messages

        #add latest 10 conversations to prompt
        for _ in range(10):
            parent_message = self.get_parent_messsage(conversation_id, parent_message_id)
            assert parent_message
            total_length += len(parent_message.message) + len(parent_message.completion)
            if total_length > 2048:
                break
            parent_messages.append(parent_message)
            parent_message_id = parent_message.parent_message_id
            if not parent_message_id:
                break

        parent_messages.reverse()
        for parent_message in parent_messages:
            context_messages.append({"role": "user", "content": parent_message.message})
            context_messages.append({"role": "assistant", "content": parent_message.completion})

        context_messages.append({"role": "user", "content": message})
        return context_messages

    async def send_message(self, conversation_id: str, message: str):
        async with self.lock:
            if self.stream:
                async for msg in self._send_message_stream(conversation_id, message):
                    yield msg
            else:
                async for msg in self._send_message(conversation_id, message):
                    yield msg

    async def _send_message(self, conversation_id: str, message: str):
        if len(message) == 0:
            return
        self.users[conversation_id] = True

        prompt = self.generate_prompt(conversation_id, message)
        logger.info('+++prompt:%s', prompt)
        try:
            yield '[BEGIN]'
            response = await openai.ChatCompletion.acreate(
                model="gpt-3.5-turbo",
                messages=prompt
            )
        except openai.error.InvalidRequestError as e:
            logger.exception(e)
            yield 'Sorry, I am not available now.'
            return
        reply = response['choices'][0]['message']['content']
        logger.info('++++response: %s', reply)
        self.add_messsage(conversation_id, message, reply)
        yield reply
        return

    async def _send_message_stream(self, conversation_id: str, message: str):
        if len(message) == 0:
            return
        self.users[conversation_id] = True

        prompt = self.generate_prompt(conversation_id, message)
        start_time = time.time()
        try:
            yield '[BEGIN]'
            response = await openai.ChatCompletion.acreate(
                model="gpt-3.5-turbo",
                messages=prompt,
                stream=True
            )
        except openai.error.InvalidRequestError as e:
            logger.exception(e)
            yield 'Sorry, I am not available now.'
            return
        collected_events = []
        completion_text = ''
        tokens: List[str] = []

        async for event in response:
            collected_events.append(event)  # save the event response
            # logger.info(event)
            delta = event['choices'][0]['delta']
            if not delta:
                break
            if not 'content' in delta:
                continue
            event_text = delta['content']  # extract the text
            tokens.append(event_text)
            if event_text.endswith('\n'):
                if time.time() - start_time > 3.0:
                    start_time = time.time()
                    reply = ''.join(tokens)
                    reply = reply.strip()
                    if reply:
                        yield reply
                    tokens = []
            completion_text += event_text  # append the text
        reply = completion_text
        logger.info('++++response: %s', reply)
        self.add_messsage(conversation_id, message, reply)
        yield ''.join(tokens)
        return




import os
from distutils.core import setup

setup(
    name="chatgpt-telegram",
    version="0.1.3",
    description="ChatGPT Bot For telegram",
    author='learnforpractice',
    license="Apache 2.0",
    url="https://github.com/learnforpractice/chatgpt-telegram",
    packages=['chatgpt_telegram'],
    package_dir={'chatgpt_telegram': 'src'},
    package_data={},
    setup_requires=['wheel']
)



# -*- coding: utf-8 -*-

import os
import sys
import time
import asyncio
import base64
import logging
import yaml
import httpx
import traceback
import websockets
import platform
from datetime import datetime
from typing import Optional, List, Dict, Any, Union, Set

import telegram
from telegram import Update
from telegram.ext import ApplicationBuilder, ContextTypes, CommandHandler, MessageHandler, filters

from dataclasses import dataclass
from playwright.async_api import async_playwright

from .chatgpt import ChatGPTBot, get_logger

logger = get_logger(__name__)

class SavedQuestion:
    def __init__(self, update: Update, context: ContextTypes.DEFAULT_TYPE, message: str = ''):
        self.update = update
        self.context = context
        self.message = message

class TelegramBot:
    def __init__(self, config_file):
        f = open(config_file)
        config = yaml.safe_load(f)

        self.telegram_api_key = config['telegram_api_key']
        self.chatgpt_accounts = config['accounts']
        self.openai_api_keys = config['openai_api_keys']
        self.tasks: List[SavedQuestion] = []
        self.saved_questions: Dict[str, SavedQuestion] = {}

        self.developer_conversation_id = None
        self.developer_user_id = None

        if 'developer_conversation_id' in config:
            self.developer_conversation_id = config['developer_conversation_id']
            self.developer_user_id = config['developer_user_id']

        self.bots: List[ChatGPTBot] = []
        self._paused = False

        self.application = ApplicationBuilder().token(self.telegram_api_key).build()    
        start_handler = CommandHandler('start', self.start)
        self.application.add_handler(start_handler)

        handler = CommandHandler('web', self.search_web)
        self.application.add_handler(handler)

        self.application.add_handler(MessageHandler(filters.TEXT & ~filters.COMMAND, self.on_message))

    async def start(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        if not self.bots:
            await self.init()
        await context.bot.send_message(chat_id=update.effective_chat.id, text="Done!")

    async def search_web(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        logger.info("+++++++search_web")
        if not self.bots:
            await self.init()
        try:
            prompt = update.message.text
            prompt = prompt.replace("/web ", "", 1)
            prompt = await self.get_web_result(prompt)
            logger.info(prompt)
            await update.message.reply_text(prompt)
            chat_type = update.message.chat.type
            if chat_type == "private":
                asyncio.create_task(self.handle_private_message(update, context, prompt))
            else:
                asyncio.create_task(self.handle_super_group_message(update, context, prompt))
        except Exception as e:
            logger.error(e)

    async def on_message(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        if not update.message:
            return

        if not self.bots:
            await self.init()

        logger.info("++++++++update.message.chat.type: %s", update.message.chat.type)
        chat_type = update.message.chat.type
        if chat_type == "private":
            asyncio.create_task(self.handle_private_message(update, context))
        # elif chat_type == "supergroup":
        else:
            asyncio.create_task(self.handle_super_group_message(update, context))

    @property
    def paused(self):
        return self._paused
    
    @paused.setter
    def paused(self, value):
        self._paused = value

    async def init(self):
        asyncio.create_task(self.handle_questions())
        if self.chatgpt_accounts:
            PLAY = await async_playwright().start()
            for account in self.chatgpt_accounts:
                user = account['user']
                psw = account['psw']
                bot = ChatGPTBot(PLAY, user, psw)
                await bot.init()
                self.bots.append(bot)

        if self.openai_api_keys:
            from .chatgpt_openai import ChatGPTBot
            for key in self.openai_api_keys:
                bot = ChatGPTBot(key)
                await bot.init()
                self.bots.append(bot)

    def choose_bot(self, user_id) -> Optional[ChatGPTBot]:
        bots = []
        for bot in self.bots:
            if bot.standby:
                continue
            bots.append(bot)

        if not bots:
            return None

        for bot in bots:
            try:
                user = bot.users[user_id]
                return bot
            except KeyError:
                pass

        bot_index = 0
        user_counts = [len(bot.users) for bot in bots]
        try:
            bot_index = user_counts.index(min(user_counts))
            logger.info("++++++choose bot: %s", bots[bot_index])
            return bots[bot_index]
        except ValueError:
            return None

    async def get_web_result(self, message: str):
        date = datetime.now()
        formatted_date = date.strftime('%m/%d/%Y')
        prompt = message
        search = message
        if message.count('/p ') == 1:
            search, prompt = message.split('/p ')
        logger.info("+++++%s %s", search, prompt)
        url = f'https://ddg-webapp-aagd.vercel.app/search?max_results=3&q="{search}"'
        r = httpx.get(url)
        results: List[Any] = r.json()
        logger.info("++++++results: %s", results)
        if not results:
            return prompt
        counter = 0
        querys = []
        querys.append("Web search results:\n\n")
        for a in results:
            counter += 1
            body = a['body']
            href = a['href']
            querys.append(f'[{counter}] "{body}"')
            querys.append(f"Source: {href}")
        querys.append(f"\nCurrent date: {formatted_date}")
        querys.append(f"\nInstructions: Using the provided web search results, write a comprehensive reply to the given prompt. Make sure to cite results using [[number](URL)] notation after the reference. If the provided search results refer to multiple subjects with the same name, write separate answers for each subject.\nPrompt: {prompt}")
        return "\n".join(querys)

    # async def echo(self, conversation_id: str, user_id: str, message: str):
    async def echo(self, update: Update, context: ContextTypes.DEFAULT_TYPE, message: str = ""):
        user_id = str(update.effective_user.id)
        if not message:
            message = update.message.text
        bot = self.choose_bot(user_id)
        if not bot:
            logger.info('no available bot')
            self.save_question(update, context)
            #queue message
            return False
        try:
            async for msg in bot.send_message(user_id, message):
                for _ in range(3):
                    try:
                        await update.message.reply_text(msg)
                        break
                    except Exception as e:
                        logger.exception(e)
            await update.message.reply_text("[END]")
            return True
        except Exception as e:
            logger.exception(e)
        self.save_question(update, context, message)
        return False

    async def echo_supergroup(self, update: Update, context: ContextTypes.DEFAULT_TYPE, message: str = ""):
        user_id = str(update.effective_user.id)
        if not message:
            message = update.message.text
        bot = self.choose_bot(user_id)
        if not bot:
            logger.info('no available bot')
            self.save_question(update, context)
            #queue message
            return False
        try:
            msgs = []
            async for msg in bot.send_message(user_id, message):
                if msg == '[BEGIN]\n':
                    await self.application.bot.send_chat_action(update.effective_chat.id, "typing")
                    continue
                msgs.append(msg)
            await update.message.reply_text(''.join(msgs))
            return True
        except Exception as e:
            logger.exception(e)
        self.save_question(update, context)
        return False

    async def handle_questions(self):
        while True:
            await asyncio.sleep(15.0)
            handled_question = []
            saved_questions = self.saved_questions.copy()
            for user_id, question in saved_questions.items():
                try:
                    user_id = str(question.update.effective_user.id)
                    message = question.message
                    logger.info("++++++++handle question: %s", message)
                    bot = self.choose_bot(user_id)
                    if not bot:
                        logger.info('no available bot')
                        break
                    msgs: List[str] = []
                    async for msg in bot.send_message(user_id, message):
                        if msg == '[BEGIN]\n':
                            await self.application.bot.send_chat_action(question.update.effective_chat.id, "typing")
                        else:
                            msgs.append(msg)
                    await question.update.message.reply_text(''.join(msgs))
                    handled_question.append(user_id)
                except Exception as e:
                    logger.exception(e)
                    continue
            for question in handled_question:
                del self.saved_questions[question]

    def save_question(self, update: Update, context: ContextTypes.DEFAULT_TYPE, message: str=''):
        user_id = str(update.effective_user.id)
        self.saved_questions[user_id] = SavedQuestion(update, context, message)

    async def handle_private_message(self, update: Update, context: ContextTypes.DEFAULT_TYPE, message: str = ""):
        try:
            await self.echo(update, context, message)
        except Exception as e:
            logger.exception(e)
            if self.developer_user_id:
                await self.sendUserText(self.developer_conversation_id, self.developer_user_id, f"exception occur at:{time.time()}: {traceback.format_exc()}")

    async def handle_super_group_message(self, update: Update, context: ContextTypes.DEFAULT_TYPE, message: str = ""):
        try:
            await self.echo_supergroup(update, context, message)
        except Exception as e:
            logger.exception(e)
            if self.developer_user_id:
                await self.sendUserText(self.developer_conversation_id, self.developer_user_id, f"exception occur at:{time.time()}: {traceback.format_exc()}")

    async def close(self):
        for bot in self.bots:
            await bot.close()

    def run(self):
        self.application.run_polling()

bot: Optional[TelegramBot]  = None

async def get_bot():
    global bot
    if not bot:
        bot = TelegramBot(sys.argv[1])
    if not bot.bots:
        await bot.init()
    return bot

async def resume():
    bot = await get_bot()
    bot.paused = False
    while not bot.paused:
        await asyncio.sleep(1.0)

def run():
    global bot
    logger.info('++++++pid: %s', os.getpid())
    logger.info('send `/start` to the bot with telegram to initialize chatgpt')
    if len(sys.argv) < 2:
        if platform.system() == 'Windows':
            print("usage: python -m chatgpt_telegram config_file")
        else:
            print("usage: python3 -m chatgpt_telegram config_file")
        return

    bot = TelegramBot(sys.argv[1])
    bot.run()

if __name__ == '__main__':
    run()
